from __future__ import print_function

import numpy as np
import tensorflow as tf
from tensorflow.contrib import rnn
import random
import collections
import time
import pickle
import datetime as dt

from scipy import linalg as LA


corpus_file='lm_corpus.pkl'
dict_file ='lm_dict.pkl'
embed_file ='lm_embeddings.pkl'

###############################################
# Load the pickle files and process data
################################################
def build_dataset():

    with open(corpus_file, 'rb') as f:
        lm_corpus=pickle.load(f)

    with open(dict_file, 'rb') as f:
        lm_dict = pickle.load(f)

    with open(embed_file, 'rb') as f:
        lm_embeds = pickle.load(f)

    reverse_lm_dict = dict(zip(lm_dict.values(), lm_dict.keys()))

    return lm_corpus, lm_dict, reverse_lm_dict, lm_embeds

########################################################
# Reduce the dimensions of embedding vectors using  PCA
########################################################
def reduce_dim(embeds, new_dim):
    org_dim=len(embeds[0])

    if (new_dim >= org_dim):
        print("New dimension should be less than the original dimension")
        return

    np_emb=np.array(embeds)
    np_emb -= np.mean(np_emb, axis=0)
    cov_emb = np.cov(np_emb, rowvar=False)

    eig_vals, eig_vecs = LA.eigh(cov_emb)

    idx = np.argsort(eig_vals)[::-1]
    eig_vecs = eig_vecs[:, idx]

    a = np.dot(np_emb, eig_vecs[:, :new_dim])
    return a



reduced_embedding_dim=5
########################################################
# Build dataset by reading the files saved in word embedding part
########################################################
lmc, lmd, rlmd, lme=build_dataset()
lme_r=reduce_dim(lme, reduced_embedding_dim)
training_data =np.array(lmc)
dictionary=lmd
reverse_dictionary=rlmd

vocab_size = len(lmd)


##############################
# Parameters
############################
learning_rate = 0.001
training_iters = 50000
display_step = 1000
time_steps = 4
input_fea_dim= 1

##########################################
# number of units in RNN cell and cell type
##########################################
n_hidden = 512
rnn_cell_type='single'

###########################################
# tf Graph input placeholders
##############################################
x = tf.placeholder("float", [None, time_steps, input_fea_dim])
y = tf.placeholder("float", [None, vocab_size])

####################################################
# Create RNN output node weights and biases
###################################################
weights = {
    'out': tf.Variable(tf.random_normal([n_hidden, vocab_size]))
}
biases = {
    'out': tf.Variable(tf.random_normal([vocab_size]))
}

############################################
#  Define the graph
##############################################
def RNN(x, weights, biases):

    # Generate a time_steps-element sequence of inputs
    x = tf.unstack(x,time_steps,1)

    #create an rnn cell

    if (rnn_cell_type is 'single'):
        rnn_cell = rnn.BasicLSTMCell(n_hidden)
    elif (rnn_cell_type is 'multi'):
        pass


    # generate prediction
    outputs, states = rnn.static_rnn(rnn_cell, x, dtype=tf.float32)

    # there are time_steps outputs but
    # we only want the last output
    return tf.matmul(outputs[-1], weights['out']) + biases['out']

##############################################################
# Create the computational graph by specifying the operations
#############################################################
pred = RNN(x, weights, biases)

# Loss and optimizer
cost = tf.reduce_mean(tf.nn.softmax_cross_entropy_with_logits(logits=pred, labels=y))
optimizer = tf.train.RMSPropOptimizer(learning_rate=learning_rate).minimize(cost)

# Model evaluation
correct_pred = tf.equal(tf.argmax(pred,1), tf.argmax(y,1))
accuracy = tf.reduce_mean(tf.cast(correct_pred, tf.float32))

# Initializing the variables
init = tf.global_variables_initializer()

############################################
# Run the graph
##############################################
with tf.Session() as session:
    session.run(init)
    step = 0
    offset = random.randint(0,time_steps+1)
    end_offset = time_steps + 1
    acc_total = 0
    loss_total = 0


    for  step in range(training_iters):
        # Generate a minibatch. Add some randomness on selection process.
        if offset > (len(training_data)-end_offset):
            offset = random.randint(0, time_steps+1)


        # values for populating placeholder x
        symbols_in_keys = [ [dictionary[ str(training_data[i])]] for i in range(offset, offset+time_steps) ]
        if (input_fea_dim==1):
            symbols_in_keys = np.reshape(np.array(symbols_in_keys), [-1, time_steps, input_fea_dim])
        else:
            symbols_in_keys= np.reshape(lme_r[symbols_in_keys,:], [-1, time_steps, input_fea_dim])


        #values for populating placeholder y
        symbols_out_onehot = np.zeros([vocab_size], dtype=float)
        symbols_out_onehot[dictionary[str(training_data[offset+time_steps])]] = 1.0
        symbols_out_onehot = np.reshape(symbols_out_onehot,[1,-1])

        #Calls session run
        _, acc, loss, onehot_pred = session.run([optimizer, accuracy, cost, pred], \
                                                feed_dict={x: symbols_in_keys, y: symbols_out_onehot})

        #update loss and accuracy
        loss_total += loss
        acc_total += acc

        #Display how the network is doing after this step
        if (step+1) % display_step == 0:
            print("Iter= " + str(step+1) + ", Average Loss= " + \
                  "{:.6f}".format(loss_total/display_step) + ", Average Accuracy= " + \
                  "{:.2f}%".format(100*acc_total/display_step))
            acc_total = 0
            loss_total = 0
            symbols_in = [training_data[i] for i in range(offset, offset + time_steps)]
            symbols_out = training_data[offset + time_steps]
            symbols_out_pred = reverse_dictionary[int(tf.argmax(onehot_pred, 1).eval())]
            print("%s - [%s] vs [%s]" % (symbols_in,symbols_out,symbols_out_pred))

        offset += (time_steps+1)
    print("Optimization Finished!")





